import numpy as np
import pandas as pd
import matplotlib.pyplot as mpl


def string_splitter(string):
    i = 0
    code = []
    while i+4 <= len(string):
        code.append(string[i:i+4])
        i += 1
    return code


def multiply_matrices(matrix_1, matrix_2):
    new_matrix = np.zeros(2, 2, 24, 24)
    new_matrix[0][0] = np.matmul(
        matrix_1[0][0], matrix_2[0][0]) + np.matmul(matrix_1[0][1], matrix_2[1][0])
    new_matrix[0][1] = np.matmul(
        matrix_1[0][0], matrix_2[0][1]) + np.matmul(matrix_1[0][1], matrix_2[1][1])
    new_matrix[1][0] = np.matmul(
        matrix_1[1][0], matrix_2[0][0]) + np.matmul(matrix_1[1][1], matrix_2[1][0])
    new_matrix[1][1] = np.matmul(
        matrix_1[1][0], matrix_2[0][1]) + np.matmul(matrix_1[1][1], matrix_2[1][1])
    return new_matrix


def instantiate_matrix_energy(K, dev, F):
    E = 0.5*(np.dot(np.dot(dev, K), dev)) + F
    return E


class polymerization:
    def __init__(self, sequence, datapath):
        self.sequence = sequence
        self.triplets = string_splitter(sequence)
        self.matrices = []
        self.eq = []
        self.fe = []
        self.datapath = datapath
        self.I_m = []
        self.twist = []
        self.roll = []
        self.tilt = []
        self.shift = []
        self.slide = []
        self.rise = []
        self.shear = []
        self.stretch = []
        self.stagger = []
        self.buckle = []
        self.propel = []
        self.opening = []
        self.var = [self.twist, self.roll, self.tilt, self.shift, self.slide, self.rise,
                    self.shear, self.stretch, self.stagger, self.buckle, self.propel, self.opening]
        self.mean_str = []

    def read_in(self):
        self.eq = []
        self.matrices = []
        tetra_mat = []
        for idx, val in enumerate(self.triplets):
            k_m = np.empty((2, 2, 24, 24))
            par = []
            fe = []
            m_00 = np.zeros((24, 24))
            m_01 = np.zeros((24, 24))
            m_10 = np.zeros((24, 24))
            m_11 = np.zeros((24, 24))

            if (len(self.triplets) > 1) and (idx < len(self.triplets)-1):
                k_m = np.empty((2, 2, 24, 24))
                par = np.empty((4, 24))
                par1 = np.zeros((24))
                par2 = np.zeros((24))
                par3 = np.zeros((24))
                par4 = np.zeros((24))
                tetra1 = val
                tetra2 = self.triplets[idx+1]
                t_1 = pd.read_csv("%s/%s_0_fe_eq.dat" %
                                    (self.datapath, tetra1), header=None, sep="\s+")
                t_1 = t_1[1]
                fe_t1 = []
                par_t1 = []
                k_t1 = pd.read_csv(
                    "%s/%s_0_fe_K.dat" % (self.datapath, tetra1), header=None, sep="\s+")
                k_t1 = k_t1.iloc[0:48, 1:49]
                k_t1 = np.asarray(k_t1)
                par1_t1 = t_1[0:24]
                par2_t1 = t_1[24:48]
                par3_t1 = t_1[48:72]
                par4_t1 = t_1[72:96]
                fe_t11 = t_1[96]
                fe_t21 = t_1[97]
                fe_t12 = t_1[98]
                fe_t22 = t_1[99]
                k_t1_00 = k_t1[0:24, 0:24]
                k_t1_01 = k_t1[0:24, 24:48]
                k_t1_10 = k_t1[24:48, 0:24]
                k_t1_11 = k_t1[24:48, 24:48]

                t_2 = pd.read_csv("%s/%s_1_fe_eq.dat" %
                                    (self.datapath, tetra1), header=None, sep="\s+")
                t_2 = t_2[1]
                fe_t2 = []
                par_t2 = []
                k_t2 = pd.read_csv(
                    "%s/%s_1_fe_K.dat" % (self.datapath, tetra1), header=None, sep="\s+")
                k_t2 = k_t2.iloc[0:48, 1:49]
                k_t2 = np.asarray(k_t2)
                par1_t2 = t_2[0:24]
                par2_t2 = t_2[24:48]
                par3_t2 = t_2[48:72]
                par4_t2 = t_2[72:96]
                fe2_t11 = t_2[96]
                fe2_t21 = t_2[97]
                fe2_t12 = t_2[98]
                fe2_t22 = t_2[99]
                k_t2_00 = k_t2[0:24, 0:24]
                k_t2_01 = k_t2[0:24, 24:48]
                k_t2_10 = k_t2[24:48, 0:24]
                k_t2_11 = k_t2[24:48, 24:48]
                mean_s = t_2[100:124]

                t_3 = pd.read_csv("%s/%s_0_fe_eq.dat" %
                                    (self.datapath, tetra2), header=None, sep="\s+")
                t_3 = t_3[1]
                fe_t3 = []
                par_t3 = []
                k_t3 = pd.read_csv(
                    "%s/%s_0_fe_K.dat" % (self.datapath, tetra2), header=None, sep="\s+")
                k_t3 = k_t3.iloc[0:48, 1:49]
                k_t3 = np.asarray(k_t3)
                par1_t3 = t_3[0:24]
                par2_t3 = t_3[24:48]
                par3_t3 = t_3[48:72]
                par4_t3 = t_3[72:96]
                fe3_t11 = t_3[96]
                fe3_t21 = t_3[97]
                fe3_t12 = t_3[98]
                fe3_t22 = t_3[99]
                k_t3_00 = k_t3[0:24, 0:24]
                k_t3_01 = k_t3[0:24, 24:48]
                k_t3_10 = k_t3[24:48, 0:24]
                k_t3_11 = k_t3[24:48, 24:48]

                t_4 = pd.read_csv("%s/%s_1_fe_eq.dat" %
                                    (self.datapath, tetra2), header=None, sep="\s+")
                t_4 = t_4[1]
                fe_t4 = []
                par_t4 = []
                k_t4 = pd.read_csv(
                    "%s/%s_1_fe_K.dat" % (self.datapath, tetra2), header=None, sep="\s+")
                k_t4 = k_t4.iloc[0:48, 1:49]
                k_t4 = np.asarray(k_t4)
                par1_t4 = t_4[0:24]
                par2_t4 = t_4[24:48]
                par3_t4 = t_4[48:72]
                par4_t4 = t_4[72:96]
                fe4_t11 = t_4[96]
                fe4_t21 = t_4[97]
                fe4_t12 = t_4[98]
                fe4_t22 = t_4[99]
                k_t4_00 = k_t4[0:24, 0:24]
                k_t4_01 = k_t4[0:24, 24:48]
                k_t4_10 = k_t4[24:48, 0:24]
                k_t4_11 = k_t4[24:48, 24:48]

                par1[0:12] = (par1_t2[0:12]+par1_t3[0:12])/2.0
                par1[12:24] = (par1_t3[12:24]+par1_t2[12:24])/2.0
                par2[0:12] = (par2_t2[0:12] + par2_t3[0:12])/2.0
                par2[12:24] = (par2_t2[12:24] + par2_t3[12:24])/2.0
                par3[0:12] = (par3_t2[0:12] + par3_t3[0:12])/2.0
                par3[12:24] = (par3_t3[12:24] + par3_t2[12:24])/2.0
                par4[0:12] = (par4_t2[0:12] + par4_t3[0:12])/2.0
                par4[12:24] = (par4_t3[12:24] + par4_t2[12:24])/2.0


                par[0] = par1
                par[1] = par2
                par[2] = par3
                par[3] = par4

                m_00 = np.zeros((24, 24))
                m_01 = np.zeros((24, 24))
                m_10 = np.zeros((24, 24))
                m_11 = np.zeros((24, 24))
                m_00[0:12, 0:12] = 1.0 * k_t2_00[0:12, 0:12] 
                m_01[0:12, 0:12] = 1.0 * k_t2_01[0:12, 0:12]
                m_10[0:12, 0:12] = 1.0 * k_t2_10[0:12, 0:12]
                m_11[0:12, 0:12] = 1.0 * k_t2_11[0:12, 0:12]

                m_00[0:12, 12:24] = (
                    k_t2_00[0:12, 12:24] + k_t3_00[0:12, 12:24])/2
                m_01[0:12, 12:24] = (
                    k_t2_01[0:12, 12:24] + k_t3_01[0:12, 12:24])/2
                m_10[0:12, 12:24] = (
                    k_t2_10[0:12, 12:24] + k_t3_10[0:12, 12:24])/2
                m_11[0:12, 12:24] = (
                    k_t2_11[0:12, 12:24] + k_t3_11[0:12, 12:24])/2

                m_00[12:24, 0:12] = (
                    k_t2_00[12:24, 0:12] + k_t3_00[12:24, 0:12])/2
                m_01[12:24, 0:12] = (
                    k_t2_01[12:24, 0:12] + k_t3_01[12:24, 0:12])/2
                m_10[12:24, 0:12] = (
                    k_t2_10[12:24, 0:12] + k_t3_10[12:24, 0:12])/2
                m_11[12:24, 0:12] = (
                    k_t2_11[12:24, 0:12] + k_t3_11[12:24, 0:12])/2

                m_00[12:24, 12:24] = 0.0 * k_t3_00[12:24, 12:24]
                m_01[12:24, 12:24] = 0.0 * k_t3_01[12:24, 12:24]
                m_10[12:24, 12:24] = 0.0 * k_t3_10[12:24, 12:24]
                m_11[12:24, 12:24] = 0.0 * k_t3_11[12:24, 12:24]

                fe11 = fe2_t11/2.0
                fe21 = fe2_t21/2.0
                fe12 = fe2_t12/2.0
                fe22 = fe2_t22/2.0
                fe.append(fe11)
                fe.append(fe12)
                fe.append(fe21)
                fe.append(fe22)

            elif len(self.triplets) > 1 and (idx == len(self.triplets)-1):
                k_m = np.empty((2, 2, 24, 24))
                par = np.empty((4, 24))
                par1 = np.zeros((24))
                par2 = np.zeros((24))
                par3 = np.zeros((24))
                par4 = np.zeros((24))
                tetra1 = val
                t_2 = pd.read_csv("%s/%s_1_fe_eq.dat" %
                                    (self.datapath, tetra1), header=None, sep="\s+")
                t_2 = t_2[1]
                fe_t2 = []
                par_t2 = []
                k_t2 = pd.read_csv(
                    "%s/%s_1_fe_K.dat" % (self.datapath, tetra1), header=None, sep="\s+")
                k_t2 = k_t2.iloc[0:48, 1:49]
                k_t2 = np.asarray(k_t2)
                par1_t2 = t_2[0:24]
                par2_t2 = t_2[24:48]
                par3_t2 = t_2[48:72]
                par4_t2 = t_2[72:96]
                fe2_t11 = t_2[96]
                fe2_t21 = t_2[97]
                fe2_t12 = t_2[98]
                fe2_t22 = t_2[99]
                k_t2_00 = k_t2[0:24, 0:24]
                k_t2_01 = k_t2[0:24, 24:48]
                k_t2_10 = k_t2[24:48, 0:24]
                k_t2_11 = k_t2[24:48, 24:48]
                mean_s = t_2[100:124]

                par1[0:12] = (par1_t2[0:12])  
                par1[12:24] = (par1_t2[12:24])  
                par2[0:12] = (par2_t2[0:12])  
                par2[12:24] = (par2_t2[12:24]) 
                par3[0:12] = (par3_t2[0:12])  
                par3[12:24] = (par3_t2[12:24]) 
                par4[0:12] = (par4_t2[0:12]) 
                par4[12:24] = (par4_t2[12:24]) 


                par[0] = par1
                par[1] = par2
                par[2] = par3
                par[3] = par4

                m_00 = np.zeros((24, 24))
                m_01 = np.zeros((24, 24))
                m_10 = np.zeros((24, 24))
                m_11 = np.zeros((24, 24))
                m_00[0:12, 0:12] = 1.0 * k_t2_00[0:12, 0:12] 
                m_01[0:12, 0:12] = 1.0 * k_t2_01[0:12, 0:12]
                m_10[0:12, 0:12] = 1.0 * k_t2_10[0:12, 0:12]
                m_11[0:12, 0:12] = 1.0 * k_t2_11[0:12, 0:12]

                m_00[0:12, 12:24] = (k_t2_00[0:12, 12:24])
                m_01[0:12, 12:24] = (k_t2_01[0:12, 12:24])
                m_10[0:12, 12:24] = (k_t2_10[0:12, 12:24])
                m_11[0:12, 12:24] = (k_t2_11[0:12, 12:24])

                m_00[12:24, 0:12] = (k_t2_00[12:24, 0:12])
                m_01[12:24, 0:12] = (k_t2_01[12:24, 0:12])
                m_10[12:24, 0:12] = (k_t2_10[12:24, 0:12])
                m_11[12:24, 0:12] = (k_t2_11[12:24, 0:12])

                m_00[12:24, 12:24] = 1.0 * k_t2_00[12:24, 12:24]
                m_01[12:24, 12:24] = 1.0 * k_t2_01[12:24, 12:24]
                m_10[12:24, 12:24] = 1.0 * k_t2_10[12:24, 12:24]
                m_11[12:24, 12:24] = 1.0 * k_t2_11[12:24, 12:24]

                fe11 = fe2_t11/2.0
                fe21 = fe2_t21/2.0
                fe12 = fe2_t12/2.0
                fe22 = fe2_t22/2.0
                fe.append(fe11)
                fe.append(fe12)
                fe.append(fe21)
                fe.append(fe22)

            elif len(self.triplets) == 1: 
                k_m = np.empty((2, 2, 24, 24))
                par = np.empty((4, 24))
                par1 = np.zeros((24))
                par2 = np.zeros((24))
                par3 = np.zeros((24))
                par4 = np.zeros((24))

                tetra1 = val
                t_2 = pd.read_csv("%s/%s_0_fe_eq.dat" %
                                    (self.datapath, tetra1), header=None, sep="\s+")
                t_2 = t_2[1]
                fe_t2 = []
                par_t2 = []
                k_t2 = pd.read_csv(
                    "%s/%s_0_fe_K.dat" % (self.datapath, tetra1), header=None, sep="\s+")
                k_t2 = k_t2.iloc[0:48, 1:49]
                k_t2 = np.asarray(k_t2)
                par1_t2 = t_2[0:24]
                par2_t2 = t_2[24:48]
                par3_t2 = t_2[48:72]
                par4_t2 = t_2[72:96]
                fe2_t11 = t_2[96]
                fe2_t21 = t_2[97]
                fe2_t12 = t_2[98]
                fe2_t22 = t_2[99]
                k_t2_00 = k_t2[0:24, 0:24]
                k_t2_01 = k_t2[0:24, 24:48]
                k_t2_10 = k_t2[24:48, 0:24]
                k_t2_11 = k_t2[24:48, 24:48]
                mean_s = t_2[100:124]
                par1[0:12] = par1_t2[0:12]
                par1[12:24] = par1_t2[12:24]
                par2[0:12] = par2_t2[0:12]
                par2[12:24] = par2_t2[12:24]
                par3[0:12] = par3_t2[0:12]
                par3[12:24] = par3_t2[12:24]
                par4[0:12] = par4_t2[0:12]
                par4[12:24] = par4_t2[12:24]

                par[0] = par1
                par[1] = par2
                par[2] = par3
                par[3] = par4
                m_00 = np.zeros((24, 24))
                m_01 = np.zeros((24, 24))
                m_10 = np.zeros((24, 24))
                m_11 = np.zeros((24, 24))
                m_00[0:12, 0:12] = 1.0 * k_t2_00[0:12, 0:12] 
                m_01[0:12, 0:12] = 1.0 * k_t2_01[0:12, 0:12]
                m_10[0:12, 0:12] = 1.0 * k_t2_10[0:12, 0:12]
                m_11[0:12, 0:12] = 1.0 * k_t2_11[0:12, 0:12]

                m_00[0:12, 12:24] = k_t2_00[0:12, 12:24]
                m_01[0:12, 12:24] = k_t2_01[0:12, 12:24]
                m_10[0:12, 12:24] = k_t2_10[0:12, 12:24]
                m_11[0:12, 12:24] = k_t2_11[0:12, 12:24]

                m_00[12:24, 0:12] = k_t2_00[12:24, 0:12]
                m_01[12:24, 0:12] = k_t2_01[12:24, 0:12]
                m_10[12:24, 0:12] = k_t2_10[12:24, 0:12]
                m_11[12:24, 0:12] = k_t2_11[12:24, 0:12]

                m_00[12:24, 12:24] = 1.0 * k_t2_00[12:24, 12:24]
                m_01[12:24, 12:24] = 1.0 * k_t2_01[12:24, 12:24]
                m_10[12:24, 12:24] = 1.0 * k_t2_10[12:24, 12:24]
                m_11[12:24, 12:24] = 1.0 * k_t2_11[12:24, 12:24]

                fe11 = fe2_t11
                fe21 = fe2_t21
                fe12 = fe2_t12
                fe22 = fe2_t22
                fe.append(fe11)
                fe.append(fe12)
                fe.append(fe21)
                fe.append(fe22)

            k_m[0][0], k_m[0][1], k_m[1][0], k_m[1][1] = m_00, m_01, m_10, m_11
            self.matrices.append(k_m)
            self.eq.append(par)
            self.fe.append(fe)
            self.mean_str.append(np.asarray(mean_s))  # mean structure

    def read_target_coordinates(self, target_path, start_step):
        variables = ["twist", "roll", "tilt", "shift", "slide", "rise",
                     "shear", "stretch", "stagger", "buckle", "propel", "opening"]
        order = []
        R = pd.DataFrame()
        steps = np.arange(1, len(self.triplets)+3, 1)
        for number in steps:
            for name in variables:
                nick = name+str(number)
                order.append(nick)
        for idx, val in enumerate(variables):
            indicators = []
            for i in steps:
                indicators.append(val+str(i))
            df = pd.read_csv("%s/test_md_%s.ser" %
                               (target_path, val), header=None, sep="\s+")
            df = df.iloc[:, start_step:start_step+len(self.triplets)+2]
            tmp = pd.DataFrame(data=np.asarray(df), columns=indicators)
            R = pd.concat((R, tmp), axis=1)
        self.coordinates = pd.DataFrame(R, columns=order)
        self.coordinates = np.asarray(self.coordinates)

    def eval_energy(self, coords, frame):
        self.I_m = np.identity(2)
        for idx, val in enumerate(self.triplets):
            if idx < len(self.triplets): 
                step_matrix = np.zeros((2, 2))
                k00, k01, k10, k11 = self.matrices[idx][0][0], self.matrices[
                    idx][0][1], self.matrices[idx][1][0], self.matrices[idx][1][1]
                eq00, eq10, eq01, eq11 = self.eq[idx][0], self.eq[idx][1], self.eq[idx][2], self.eq[idx][3]
                fe00, fe10, fe01, fe11 = self.fe[idx][0], self.fe[idx][1], self.fe[idx][2], self.fe[idx][3]
                fe_Min = np.min((fe00, fe10, fe01, fe11))
                fe00, fe10, fe01, fe11 = fe00 - fe_Min, fe10 - \
                    fe_Min, fe01 - fe_Min, fe11 - fe_Min
                coord_triplet = coords[idx]
                dev00, dev10, dev01, dev11 = coord_triplet-eq00, coord_triplet - \
                    eq10, coord_triplet-eq01, coord_triplet-eq11
                E_00 = instantiate_matrix_energy(k00, dev00, fe00)
                E_10 = instantiate_matrix_energy(k10, dev10, fe10)
                E_01 = instantiate_matrix_energy(k01, dev01, fe01)
                E_11 = instantiate_matrix_energy(k11, dev11, fe11)
                Es = [E_00, E_10, E_01, E_11]
                Em = np.min(Es)
                imin = np.argwhere(Es == Em)
                step_matrix[0][0] = E_00 
                step_matrix[1][0] = E_10
                step_matrix[0][1] = E_01
                step_matrix[1][1] = E_11
                step_matrix = np.float128(step_matrix)
                self.I_m = np.matmul(self.I_m, np.exp(-step_matrix/0.593))
        Fe = np.sum(self.I_m)
        return -0.593*np.log(Fe)

    def run_Ising_trajectory(self, target_path, start_step, frame_number):
        self.read_in()
        self.read_target_coordinates(target_path, start_step)
        frame = 0
        energy = []
        while frame <= frame_number:
            coords = []
            i = 0 
            while i < len(self.triplets): 
                co = self.coordinates[frame][(i)*12:(i+2)*12]
                coords.append(co)
                i += 1
            FE = self.eval_energy(coords, frame)
            energy.append(FE)
            frame += 1
        FE = FE - np.min(FE)
        return np.asarray(energy)
